@@ -1,5 +1,4 @@
 """Helper functions for a standard streaming compression API"""
-from bz2 import BZ2File
 from zipfile import ZipFile
 
 import fsspec.utils
@@ -68,17 +67,21 @@ def unzip(infile, mode="rb", filename=None, **kwargs):
 
 
 register_compression("zip", unzip, "zip")
-register_compression("bz2", BZ2File, "bz2")
+
+try:
+    from bz2 import BZ2File
+except ImportError:
+    pass
+else:
+    register_compression("bz2", BZ2File, "bz2")
 
 try:  # pragma: no cover
     from isal import igzip
 
-    # igzip is meant to be used as a faster drop in replacement to gzip
-    # so its api and functions are the same as the stdlib’s module. Except
-    # where ISA-L does not support the same calls as zlib
-    # (See https://python-isal.readthedocs.io/).
+    def isal(infile, mode="rb", **kwargs):
+        return igzip.IGzipFile(fileobj=infile, mode=mode, **kwargs)
 
-    register_compression("gzip", igzip.IGzipFile, "gz")
+    register_compression("gzip", isal, "gz")
 except ImportError:
     from gzip import GzipFile
 
@@ -142,7 +145,7 @@ try:
     # standard implementation.
     register_compression("snappy", SnappyFile, [])
 
-except (ImportError, NameError):
+except (ImportError, NameError, AttributeError):
     pass
 
 try:
