@@ -1,5 +1,3 @@
-from __future__ import absolute_import, division, print_function
-
 import io
 import logging
 import os
@@ -18,18 +16,16 @@ from .caching import (  # noqa: F401
 from .compression import compr
 from .registry import filesystem, get_filesystem_class
 from .utils import (
-    IOWrapper,
     _unstrip_protocol,
     build_name_function,
     infer_compression,
     stringify_path,
-    update_storage_options,
 )
 
 logger = logging.getLogger("fsspec")
 
 
-class OpenFile(object):
+class OpenFile:
     """
     File-like object to be used in a context
 
@@ -37,7 +33,7 @@ class OpenFile(object):
     are typically binary-only.
 
     These instances are safe to serialize, as the low-level file object
-    is not created until invoked using `with`.
+    is not created until invoked using ``with``.
 
     Parameters
     ----------
@@ -56,6 +52,10 @@ class OpenFile(object):
         How to handle encoding errors if opened in text mode.
     newline: None or str
         Passed to TextIOWrapper in text mode, how to handle line endings.
+    autoopen: bool
+        If True, calls open() immediately. Mostly used by pickle
+    pos: int
+        If given and autoopen is True, seek to this location immediately
     """
 
     def __init__(
@@ -94,10 +94,6 @@ class OpenFile(object):
     def __repr__(self):
         return "<OpenFile '{}'>".format(self.path)
 
-    def __fspath__(self):
-        # may raise if cannot be resolved to local file
-        return self.open().__fspath__()
-
     def __enter__(self):
         mode = self.mode.replace("t", "").replace("b", "") + "b"
 
@@ -112,7 +108,7 @@ class OpenFile(object):
 
         if "b" not in self.mode:
             # assume, for example, that 'r' is equivalent to 'rt' as in builtin
-            f = io.TextIOWrapper(
+            f = PickleableTextIOWrapper(
                 f, encoding=self.encoding, errors=self.errors, newline=self.newline
             )
             self.fobjects.append(f)
@@ -122,10 +118,6 @@ class OpenFile(object):
     def __exit__(self, *args):
         self.close()
 
-    def __del__(self):
-        if hasattr(self, "fobjects"):
-            self.fobjects.clear()  # may cause cleanup of objects and close files
-
     @property
     def full_name(self):
         return _unstrip_protocol(self.path, self.fs)
@@ -133,31 +125,19 @@ class OpenFile(object):
     def open(self):
         """Materialise this as a real open file without context
 
-        The file should be explicitly closed to avoid enclosed file
-        instances persisting. This code-path monkey-patches the file-like
-        objects, so they can close even if the parent OpenFile object has already
-        been deleted; but a with-context is better style.
+        The OpenFile object should be explicitly closed to avoid enclosed file
+        instances persisting. You must, therefore, keep a reference to the OpenFile
+        during the life of the file-like it generates.
         """
-        out = self.__enter__()
-        closer = out.close
-        fobjects = self.fobjects.copy()[:-1]
-        mode = self.mode
-
-        def close():
-            # this func has no reference to
-            closer()  # original close bound method of the final file-like
-            _close(fobjects, mode)  # call close on other dependent file-likes
-
-        try:
-            out.close = close
-        except AttributeError:
-            out = IOWrapper(out, lambda: _close(fobjects, mode))
-
-        return out
+        return self.__enter__()
 
     def close(self):
         """Close all encapsulated file objects"""
-        _close(self.fobjects, self.mode)
+        for f in reversed(self.fobjects):
+            if "r" not in self.mode and not f.closed:
+                f.flush()
+            f.close()
+        self.fobjects.clear()
 
 
 class OpenFiles(list):
@@ -196,18 +176,17 @@ class OpenFiles(list):
 
     def __exit__(self, *args):
         fs = self.fs
+        [s.__exit__(*args) for s in self]
         if "r" not in self.mode:
             while True:
                 if hasattr(fs, "open_many"):
                     # check for concurrent cache upload
                     fs.commit_many(self.files)
-                    self.files.clear()
                     return
                 if hasattr(fs, "fs") and fs.fs is not None:
                     fs = fs.fs
                 else:
                     break
-        [s.__exit__(*args) for s in self]
 
     def __getitem__(self, item):
         out = super().__getitem__(item)
@@ -219,14 +198,6 @@ class OpenFiles(list):
         return "<List of %s OpenFile instances>" % len(self)
 
 
-def _close(fobjects, mode):
-    for f in reversed(fobjects):
-        if "r" not in mode and not f.closed:
-            f.flush()
-        f.close()
-    fobjects.clear()
-
-
 def open_files(
     urlpath,
     mode="rb",
@@ -295,6 +266,16 @@ def open_files(
     -------
     An ``OpenFiles`` instance, which is a list of ``OpenFile`` objects that can
     be used as a single context
+
+    Notes
+    -----
+    For a full list of the available protocols and the implementations that
+    they map across to see the latest online documentation:
+
+    - For implementations built into ``fsspec`` see
+      https://filesystem-spec.readthedocs.io/en/latest/api.html#built-in-implementations
+    - For implementations in separate packages see
+      https://filesystem-spec.readthedocs.io/en/latest/api.html#other-known-implementations
     """
     fs, fs_token, paths = get_fs_token_paths(
         urlpath,
@@ -305,7 +286,9 @@ def open_files(
         protocol=protocol,
         expand=expand,
     )
-    if "r" not in mode and auto_mkdir:
+    if fs.protocol == "file":
+        fs.auto_mkdir = auto_mkdir
+    elif "r" not in mode and auto_mkdir:
         parents = {fs._parent(path) for path in paths}
         [fs.makedirs(parent, exist_ok=True) for parent in parents]
     return OpenFiles(
@@ -327,35 +310,23 @@ def open_files(
 
 
 def _un_chain(path, kwargs):
-    if isinstance(path, (tuple, list)):
-        bits = [_un_chain(p, kwargs) for p in path]
-        out = []
-        for pbit in zip(*bits):
-            paths, protocols, kwargs = zip(*pbit)
-            if len(set(protocols)) > 1:
-                raise ValueError("Protocol mismatch in URL chain")
-            if len(set(paths)) == 1:
-                paths = paths[0]
-            else:
-                paths = list(paths)
-            out.append([paths, protocols[0], kwargs[0]])
-        return out
     x = re.compile(".*[^a-z]+.*")  # test for non protocol-like single word
     bits = (
         [p if "://" in p or x.match(p) else p + "://" for p in path.split("::")]
         if "::" in path
         else [path]
     )
-    if len(bits) < 2:
-        return []
     # [[url, protocol, kwargs], ...]
     out = []
     previous_bit = None
+    kwargs = kwargs.copy()
     for bit in reversed(bits):
-        protocol = split_protocol(bit)[0] or "file"
+        protocol = kwargs.pop("protocol", None) or split_protocol(bit)[0] or "file"
         cls = get_filesystem_class(protocol)
         extra_kwargs = cls._get_kwargs_from_urls(bit)
-        kws = kwargs.get(protocol, {})
+        kws = kwargs.pop(protocol, {})
+        if bit is bits[0]:
+            kws.update(kwargs)
         kw = dict(**extra_kwargs, **kws)
         bit = cls._strip_protocol(bit)
         if (
@@ -389,28 +360,32 @@ def url_to_fs(url, **kwargs):
     urlpath : str
         The file-systems-specific URL for ``url``.
     """
+    # non-FS arguments that appear in fsspec.open()
+    # inspect could keep this in sync with open()'s signature
+    known_kwargs = {
+        "compression",
+        "encoding",
+        "errors",
+        "expand",
+        "mode",
+        "name_function",
+        "newline",
+        "num",
+    }
+    kwargs = {k: v for k, v in kwargs.items() if k not in known_kwargs}
     chain = _un_chain(url, kwargs)
-    if len(chain) > 1:
-        inkwargs = {}
-        # Reverse iterate the chain, creating a nested target_* structure
-        for i, ch in enumerate(reversed(chain)):
-            urls, protocol, kw = ch
-            if i == len(chain) - 1:
-                inkwargs = dict(**kw, **inkwargs)
-                continue
-            inkwargs["target_options"] = dict(**kw, **inkwargs)
-            inkwargs["target_protocol"] = protocol
-            inkwargs["fo"] = urls
-        urlpath, protocol, _ = chain[0]
-        fs = filesystem(protocol, **inkwargs)
-    else:
-        protocol = split_protocol(url)[0]
-        cls = get_filesystem_class(protocol)
-
-        options = cls._get_kwargs_from_urls(url)
-        update_storage_options(options, kwargs)
-        fs = cls(**options)
-        urlpath = fs._strip_protocol(url)
+    inkwargs = {}
+    # Reverse iterate the chain, creating a nested target_* structure
+    for i, ch in enumerate(reversed(chain)):
+        urls, protocol, kw = ch
+        if i == len(chain) - 1:
+            inkwargs = dict(**kw, **inkwargs)
+            continue
+        inkwargs["target_options"] = dict(**kw, **inkwargs)
+        inkwargs["target_protocol"] = protocol
+        inkwargs["fo"] = urls
+    urlpath, protocol, _ = chain[0]
+    fs = filesystem(protocol, **inkwargs)
     return fs, urlpath
 
 
@@ -463,8 +438,18 @@ def open(
     Returns
     -------
     ``OpenFile`` object.
+
+    Notes
+    -----
+    For a full list of the available protocols and the implementations that
+    they map across to see the latest online documentation:
+
+    - For implementations built into ``fsspec`` see
+      https://filesystem-spec.readthedocs.io/en/latest/api.html#built-in-implementations
+    - For implementations in separate packages see
+      https://filesystem-spec.readthedocs.io/en/latest/api.html#other-known-implementations
     """
-    return open_files(
+    out = open_files(
         urlpath=[urlpath],
         mode=mode,
         compression=compression,
@@ -474,7 +459,10 @@ def open(
         newline=newline,
         expand=False,
         **kwargs,
-    )[0]
+    )
+    if not out:
+        raise FileNotFoundError(urlpath)
+    return out[0]
 
 
 def open_local(url, mode="rb", **storage_options):
@@ -613,62 +601,41 @@ def get_fs_token_paths(
     if isinstance(urlpath, (list, tuple, set)):
         if not urlpath:
             raise ValueError("empty urlpath sequence")
-        urlpath = [stringify_path(u) for u in urlpath]
+        urlpath0 = stringify_path(list(urlpath)[0])
     else:
-        urlpath = stringify_path(urlpath)
-    chain = _un_chain(urlpath, storage_options or {})
-    if len(chain) > 1:
-        inkwargs = {}
-        # Reverse iterate the chain, creating a nested target_* structure
-        for i, ch in enumerate(reversed(chain)):
-            urls, nested_protocol, kw = ch
-            if i == len(chain) - 1:
-                inkwargs = dict(**kw, **inkwargs)
-                continue
-            inkwargs["target_options"] = dict(**kw, **inkwargs)
-            inkwargs["target_protocol"] = nested_protocol
-            inkwargs["fo"] = urls
-        paths, protocol, _ = chain[0]
-        fs = filesystem(protocol, **inkwargs)
-        if isinstance(paths, (list, tuple, set)):
-            paths = [fs._strip_protocol(u) for u in paths]
-        else:
-            paths = fs._strip_protocol(paths)
+        urlpath0 = stringify_path(urlpath)
+    storage_options = storage_options or {}
+    if protocol:
+        storage_options["protocol"] = protocol
+    chain = _un_chain(urlpath0, storage_options or {})
+    inkwargs = {}
+    # Reverse iterate the chain, creating a nested target_* structure
+    for i, ch in enumerate(reversed(chain)):
+        urls, nested_protocol, kw = ch
+        if i == len(chain) - 1:
+            inkwargs = dict(**kw, **inkwargs)
+            continue
+        inkwargs["target_options"] = dict(**kw, **inkwargs)
+        inkwargs["target_protocol"] = nested_protocol
+        inkwargs["fo"] = urls
+    paths, protocol, _ = chain[0]
+    fs = filesystem(protocol, **inkwargs)
+    if isinstance(urlpath, (list, tuple, set)):
+        pchains = [
+            _un_chain(stringify_path(u), storage_options or {})[0] for u in urlpath
+        ]
+        if len({pc[1] for pc in pchains}) > 1:
+            raise ValueError("Protocol mismatch getting fs from %s", urlpath)
+        paths = [pc[0] for pc in pchains]
     else:
-        if isinstance(urlpath, (list, tuple, set)):
-            protocols, paths = zip(*map(split_protocol, urlpath))
-            if protocol is None:
-                protocol = protocols[0]
-                if not all(p == protocol for p in protocols):
-                    raise ValueError(
-                        "When specifying a list of paths, all paths must "
-                        "share the same protocol"
-                    )
-            cls = get_filesystem_class(protocol)
-            optionss = list(map(cls._get_kwargs_from_urls, urlpath))
-            paths = [cls._strip_protocol(u) for u in urlpath]
-            options = optionss[0]
-            if not all(o == options for o in optionss):
-                raise ValueError(
-                    "When specifying a list of paths, all paths must "
-                    "share the same file-system options"
-                )
-            update_storage_options(options, storage_options)
-            fs = cls(**options)
-        else:
-            protocols = split_protocol(urlpath)[0]
-            protocol = protocol or protocols
-            cls = get_filesystem_class(protocol)
-            options = cls._get_kwargs_from_urls(urlpath)
-            paths = cls._strip_protocol(urlpath)
-            update_storage_options(options, storage_options)
-            fs = cls(**options)
-
+        paths = fs._strip_protocol(paths)
     if isinstance(paths, (list, tuple, set)):
         paths = expand_paths_if_needed(paths, mode, num, fs, name_function)
     else:
         if "w" in mode and expand:
             paths = _expand_paths(paths, name_function, num)
+        elif "x" in mode and expand:
+            paths = _expand_paths(paths, name_function, num)
         elif "*" in paths:
             paths = [f for f in sorted(fs.glob(paths)) if not fs.isdir(f)]
         else:
@@ -705,3 +672,26 @@ def _expand_paths(path, name_function, num):
             "3. A path with a '*' in it: 'foo.*.json'"
         )
     return paths
+
+
+class PickleableTextIOWrapper(io.TextIOWrapper):
+    """TextIOWrapper cannot be pickled. This solves it.
+
+    Requires that ``buffer`` be pickleable, which all instances of
+    AbstractBufferedFile are.
+    """
+
+    def __init__(
+        self,
+        buffer,
+        encoding=None,
+        errors=None,
+        newline=None,
+        line_buffering=False,
+        write_through=False,
+    ):
+        self.args = buffer, encoding, errors, newline, line_buffering, write_through
+        super().__init__(*self.args)
+
+    def __reduce__(self):
+        return PickleableTextIOWrapper, self.args
