@@ -1,15 +1,75 @@
-from __future__ import absolute_import, division, print_function
-
 from contextlib import contextmanager
+from ctypes import (
+    CFUNCTYPE,
+    POINTER,
+    c_int,
+    c_longlong,
+    c_void_p,
+    cast,
+    create_string_buffer,
+)
 
 import libarchive
+import libarchive.ffi as ffi
 
-from fsspec import AbstractFileSystem, open_files
+from fsspec import open_files
+from fsspec.archive import AbstractArchiveFileSystem
 from fsspec.implementations.memory import MemoryFile
-from fsspec.utils import DEFAULT_BLOCK_SIZE, tokenize
+from fsspec.utils import DEFAULT_BLOCK_SIZE
+
+# Libarchive requires seekable files or memory only for certain archive
+# types. However, since we read the directory first to cache the contents
+# and also allow random access to any file, the file-like object needs
+# to be seekable no matter what.
+
+# Seek call-backs (not provided in the libarchive python wrapper)
+SEEK_CALLBACK = CFUNCTYPE(c_longlong, c_int, c_void_p, c_longlong, c_int)
+read_set_seek_callback = ffi.ffi(
+    "read_set_seek_callback", [ffi.c_archive_p, SEEK_CALLBACK], c_int, ffi.check_int
+)
+new_api = hasattr(ffi, "NO_OPEN_CB")
 
 
-class LibArchiveFileSystem(AbstractFileSystem):
+@contextmanager
+def custom_reader(file, format_name="all", filter_name="all", block_size=ffi.page_size):
+    """Read an archive from a seekable file-like object.
+
+    The `file` object must support the standard `readinto` and 'seek' methods.
+    """
+    buf = create_string_buffer(block_size)
+    buf_p = cast(buf, c_void_p)
+
+    def read_func(archive_p, context, ptrptr):
+        # readinto the buffer, returns number of bytes read
+        length = file.readinto(buf)
+        # write the address of the buffer into the pointer
+        ptrptr = cast(ptrptr, POINTER(c_void_p))
+        ptrptr[0] = buf_p
+        # tell libarchive how much data was written into the buffer
+        return length
+
+    def seek_func(archive_p, context, offset, whence):
+        file.seek(offset, whence)
+        # tell libarchvie the current position
+        return file.tell()
+
+    read_cb = ffi.READ_CALLBACK(read_func)
+    seek_cb = SEEK_CALLBACK(seek_func)
+
+    if new_api:
+        open_cb = ffi.NO_OPEN_CB
+        close_cb = ffi.NO_CLOSE_CB
+    else:
+        open_cb = libarchive.read.OPEN_CALLBACK(ffi.VOID_CB)
+        close_cb = libarchive.read.CLOSE_CALLBACK(ffi.VOID_CB)
+
+    with libarchive.read.new_archive_read(format_name, filter_name) as archive_p:
+        read_set_seek_callback(archive_p, seek_cb)
+        ffi.read_open(archive_p, None, open_cb, read_cb, close_cb)
+        yield libarchive.read.ArchiveRead(archive_p)
+
+
+class LibArchiveFileSystem(AbstractArchiveFileSystem):
     """Compressed archives as a file-system (read-only)
 
     Supports the following formats:
@@ -17,14 +77,19 @@ class LibArchiveFileSystem(AbstractFileSystem):
     Microsoft CAB, 7-Zip, WARC
 
     See the libarchive documentation for further restrictions.
+    https://www.libarchive.org/
 
-    Keeps file object open while instance lives.
+    Keeps file object open while instance lives. It only works in seekable
+    file-like objects. In case the filesystem does not support this kind of
+    file object, it is recommended to cache locally.
 
-    This class is pickleable, but not necessarily thread-safe
+    This class is pickleable, but not necessarily thread-safe (depends on the
+    platform). See libarchive documentation for details.
     """
 
     root_marker = ""
     protocol = "libarchive"
+    cachable = False
 
     def __init__(
         self,
@@ -40,7 +105,7 @@ class LibArchiveFileSystem(AbstractFileSystem):
         ----------
         fo: str or file-like
             Contains ZIP, and must exist. If a str, will fetch file using
-            `open_files()`, which must return one file exactly.
+            :meth:`~fsspec.open_files`, which must return one file exactly.
         mode: str
             Currently, only 'r' accepted
         target_protocol: str (optional)
@@ -61,15 +126,15 @@ class LibArchiveFileSystem(AbstractFileSystem):
                     'one file: "{}"'.format(fo, files)
                 )
             fo = files[0]
+        self.of = fo
         self.fo = fo.__enter__()  # the whole instance is a context
-        # self.arc_reader =
         self.block_size = block_size
         self.dir_cache = None
 
     @contextmanager
     def _open_archive(self):
         self.fo.seek(0)
-        with libarchive.fd_reader(self.fo.fileno(), block_size=self.block_size) as arc:
+        with custom_reader(self.fo, block_size=self.block_size) as arc:
             yield arc
 
     @classmethod
@@ -119,43 +184,6 @@ class LibArchiveFileSystem(AbstractFileSystem):
             }
         )
 
-    def info(self, path, **kwargs):
-        self._get_dirs()
-        path = self._strip_protocol(path)
-        if path in self.dir_cache:
-            return self.dir_cache[path]
-        elif path + "/" in self.dir_cache:
-            return self.dir_cache[path + "/"]
-        else:
-            raise FileNotFoundError(path)
-
-    def ls(self, path, detail=False, **kwargs):
-        self._get_dirs()
-        paths = {}
-
-        for p, f in self.dir_cache.items():
-            p = p.rstrip("/")
-            if "/" in p:
-                root = p.rsplit("/", 1)[0]
-            else:
-                root = ""
-            if root == path.rstrip("/"):
-                paths[p] = f
-            elif all(
-                (a == b)
-                for a, b in zip(path.split("/"), [""] + p.strip("/").split("/"))
-            ):
-                # root directory entry
-                ppath = p.rstrip("/").split("/", 1)[0]
-                if ppath not in paths:
-                    out = {"name": ppath + "/", "size": 0, "type": "directory"}
-                    paths[ppath] = out
-        out = list(paths.values())
-        if detail:
-            return out
-        else:
-            return list(sorted(f["name"] for f in out))
-
     def _open(
         self,
         path,
@@ -171,29 +199,17 @@ class LibArchiveFileSystem(AbstractFileSystem):
 
         data = bytes()
         with self._open_archive() as arc:
-            # FIXME? dropwhile would increase performance but less readable
             for entry in arc:
                 if entry.pathname != path:
                     continue
+
+                if entry.size == 0:
+                    # empty file, so there are no blocks
+                    break
+
                 for block in entry.get_blocks(entry.size):
                     data = block
                     break
                 else:
                     raise ValueError
         return MemoryFile(fs=self, path=path, data=data)
-
-    def ukey(self, path):
-        return tokenize(path, self.fo, self.protocol)
-
-    def _all_dirnames(self, paths):
-        """Returns *all* directory names for each path in paths, including intermediate ones.
-
-        Parameters
-        ----------
-        paths: Iterable of path strings
-        """
-        if len(paths) == 0:
-            return set()
-
-        dirnames = {self._parent(path) for path in paths} - {self.root_marker}
-        return dirnames | self._all_dirnames(dirnames)
