@@ -1,5 +1,6 @@
 import datetime
 import io
+import logging
 import os
 import os.path as osp
 import posixpath
@@ -13,6 +14,8 @@ from fsspec.compression import compr
 from fsspec.core import get_compression
 from fsspec.utils import isfilelike, stringify_path
 
+logger = logging.getLogger("fsspec.local")
+
 
 class LocalFileSystem(AbstractFileSystem):
     """Interface to files on local storage
@@ -33,6 +36,10 @@ class LocalFileSystem(AbstractFileSystem):
         super().__init__(**kwargs)
         self.auto_mkdir = auto_mkdir
 
+    @property
+    def fsid(self):
+        return "local"
+
     def mkdir(self, path, create_parents=True, **kwargs):
         path = self._strip_protocol(path)
         if self.exists(path):
@@ -58,10 +65,6 @@ class LocalFileSystem(AbstractFileSystem):
         else:
             return [posixpath.join(path, f) for f in os.listdir(path)]
 
-    def glob(self, path, **kwargs):
-        path = self._strip_protocol(path)
-        return super().glob(path, **kwargs)
-
     def info(self, path, **kwargs):
         if isinstance(path, os.DirEntry):
             # scandir DirEntry
@@ -94,14 +97,14 @@ class LocalFileSystem(AbstractFileSystem):
             "created": out.st_ctime,
             "islink": link,
         }
-        for field in ["mode", "uid", "gid", "mtime"]:
+        for field in ["mode", "uid", "gid", "mtime", "ino", "nlink"]:
             result[field] = getattr(out, "st_" + field)
         if result["islink"]:
             result["destination"] = os.readlink(path)
             try:
                 out2 = os.stat(path, follow_symlinks=True)
                 result["size"] = out2.st_size
-            except IOError:
+            except OSError:
                 result["size"] = 0
         return result
 
@@ -118,7 +121,7 @@ class LocalFileSystem(AbstractFileSystem):
         elif self.isdir(path1):
             self.mkdirs(path2, exist_ok=True)
         else:
-            raise FileNotFoundError
+            raise FileNotFoundError(path1)
 
     def get_file(self, path1, path2, callback=None, **kwargs):
         if isfilelike(path2):
@@ -135,30 +138,48 @@ class LocalFileSystem(AbstractFileSystem):
         path2 = self._strip_protocol(path2).rstrip("/")
         shutil.move(path1, path2)
 
+    def link(self, src, dst, **kwargs):
+        src = self._strip_protocol(src)
+        dst = self._strip_protocol(dst)
+        os.link(src, dst, **kwargs)
+
+    def symlink(self, src, dst, **kwargs):
+        src = self._strip_protocol(src)
+        dst = self._strip_protocol(dst)
+        os.symlink(src, dst, **kwargs)
+
+    def islink(self, path) -> bool:
+        return os.path.islink(self._strip_protocol(path))
+
     def rm_file(self, path):
-        os.remove(path)
+        os.remove(self._strip_protocol(path))
 
     def rm(self, path, recursive=False, maxdepth=None):
-        if isinstance(path, str):
+        if not isinstance(path, list):
             path = [path]
 
         for p in path:
             p = self._strip_protocol(p).rstrip("/")
-            if recursive and self.isdir(p):
-
+            if self.isdir(p):
+                if not recursive:
+                    raise ValueError("Cannot delete directory, set recursive=True")
                 if osp.abspath(p) == os.getcwd():
                     raise ValueError("Cannot delete current working directory")
                 shutil.rmtree(p)
             else:
                 os.remove(p)
 
+    def unstrip_protocol(self, name):
+        name = self._strip_protocol(name)  # normalise for local/win/...
+        return f"file://{name}"
+
     def _open(self, path, mode="rb", block_size=None, **kwargs):
         path = self._strip_protocol(path)
         if self.auto_mkdir and "w" in mode:
             self.makedirs(self._parent(path), exist_ok=True)
         return LocalFileOpener(path, mode, fs=self, **kwargs)
 
-    def touch(self, path, **kwargs):
+    def touch(self, path, truncate=True, **kwargs):
         path = self._strip_protocol(path)
         if self.auto_mkdir:
             self.makedirs(self._parent(path), exist_ok=True)
@@ -166,14 +187,18 @@ class LocalFileSystem(AbstractFileSystem):
             os.utime(path, None)
         else:
             open(path, "a").close()
+        if truncate:
+            os.truncate(path, 0)
 
     def created(self, path):
         info = self.info(path=path)
-        return datetime.datetime.utcfromtimestamp(info["created"])
+        return datetime.datetime.fromtimestamp(
+            info["created"], tz=datetime.timezone.utc
+        )
 
     def modified(self, path):
         info = self.info(path=path)
-        return datetime.datetime.utcfromtimestamp(info["mtime"])
+        return datetime.datetime.fromtimestamp(info["mtime"], tz=datetime.timezone.utc)
 
     @classmethod
     def _parent(cls, path):
@@ -188,6 +213,8 @@ class LocalFileSystem(AbstractFileSystem):
         path = stringify_path(path)
         if path.startswith("file://"):
             path = path[7:]
+        elif path.startswith("file:"):
+            path = path[5:]
         return make_path_posix(path).rstrip("/") or cls.root_marker
 
     def _isfilestore(self):
@@ -211,6 +238,8 @@ def make_path_posix(path, sep=os.sep):
         # most common fast case for posix
         if path.startswith("/"):
             return path
+        if path.startswith("./"):
+            path = path[2:]
         return os.getcwd() + "/" + path
     if (
         (sep not in path and "/" not in path)
@@ -223,10 +252,12 @@ def make_path_posix(path, sep=os.sep):
             return make_path_posix(osp.abspath(path))
         else:
             return os.getcwd() + "/" + path
+    if path.startswith("file://"):
+        path = path[7:]
     if re.match("/[A-Za-z]:", path):
         # for windows file URI like "file:///C:/folder/file"
         # or "file:///C:\\dir\\file"
-        path = path[1:]
+        path = path[1:].replace("\\", "/").replace("//", "/")
     if path.startswith("\\\\"):
         # special case for windows UNC/DFS-style paths, do nothing,
         # just flip the slashes around (case below does not work!)
@@ -240,10 +271,23 @@ def make_path_posix(path, sep=os.sep):
     return path
 
 
+def trailing_sep(path):
+    """Return True if the path ends with a path separator.
+
+    A forward slash is always considered a path separator, even on Operating
+    Systems that normally use a backslash.
+    """
+    # TODO: if all incoming paths were posix-compliant then separator would
+    # always be a forward slash, simplifying this function.
+    # See https://github.com/fsspec/filesystem_spec/pull/1250
+    return path.endswith(os.sep) or (os.altsep is not None and path.endswith(os.altsep))
+
+
 class LocalFileOpener(io.IOBase):
     def __init__(
         self, path, mode, autocommit=True, fs=None, compression=None, **kwargs
     ):
+        logger.debug("open file: %s", path)
         self.path = path
         self.mode = mode
         self.fs = fs
@@ -342,9 +386,11 @@ class LocalFileOpener(io.IOBase):
     def closed(self):
         return self.f.closed
 
-    def __fspath__(self):
-        # uniquely among fsspec implementations, this is a real, local path
-        return self.path
+    def fileno(self):
+        return self.raw.fileno()
+
+    def flush(self) -> None:
+        self.f.flush()
 
     def __iter__(self):
         return self.f.__iter__()
@@ -354,7 +400,7 @@ class LocalFileOpener(io.IOBase):
 
     def __enter__(self):
         self._incontext = True
-        return self.f.__enter__()
+        return self
 
     def __exit__(self, exc_type, exc_value, traceback):
         self._incontext = False
