@@ -1,5 +1,8 @@
 import array
+import posixpath
+import warnings
 from collections.abc import MutableMapping
+from functools import cached_property
 
 from .core import url_to_fs
 
@@ -34,9 +37,8 @@ class FSMap(MutableMapping):
 
     def __init__(self, root, fs, check=False, create=False, missing_exceptions=None):
         self.fs = fs
-        self.root = fs._strip_protocol(root).rstrip(
-            "/"
-        )  # we join on '/' in _key_to_str
+        self.root = fs._strip_protocol(root).rstrip("/")
+        self._root_key_to_str = fs._strip_protocol(posixpath.join(root, "x"))[:-1]
         if missing_exceptions is None:
             missing_exceptions = (
                 FileNotFoundError,
@@ -44,6 +46,8 @@ class FSMap(MutableMapping):
                 NotADirectoryError,
             )
         self.missing_exceptions = missing_exceptions
+        self.check = check
+        self.create = create
         if create:
             if not self.fs.exists(root):
                 self.fs.mkdir(root)
@@ -56,6 +60,13 @@ class FSMap(MutableMapping):
             self.fs.touch(root + "/a")
             self.fs.rm(root + "/a")
 
+    @cached_property
+    def dirfs(self):
+        """dirfs instance that can be used with the same keys as the mapper"""
+        from .implementations.dirfs import DirFileSystem
+
+        return DirFileSystem(path=self._root_key_to_str, fs=self.fs)
+
     def clear(self):
         """Remove all keys below root - empties out mapping"""
         try:
@@ -118,11 +129,16 @@ class FSMap(MutableMapping):
 
     def _key_to_str(self, key):
         """Generate full path for the key"""
-        if isinstance(key, (tuple, list)):
-            key = str(tuple(key))
-        else:
+        if not isinstance(key, str):
+            # raise TypeError("key must be of type `str`, got `{type(key).__name__}`"
+            warnings.warn(
+                "from fsspec 2023.5 onward FSMap non-str keys will raise TypeError",
+                DeprecationWarning,
+            )
+            if isinstance(key, list):
+                key = tuple(key)
             key = str(key)
-        return self.fs._strip_protocol("/".join([self.root, key]) if self.root else key)
+        return f"{self._root_key_to_str}{key}"
 
     def _str_to_key(self, s):
         """Strip path of to leave key name"""
@@ -140,6 +156,7 @@ class FSMap(MutableMapping):
         return result
 
     def pop(self, key, default=None):
+        """Pop data"""
         result = self.__getitem__(key, default)
         try:
             del self[key]
@@ -182,12 +199,12 @@ def maybe_convert(value):
             # The buffer interface doesn't support datetime64/timdelta64 numpy
             # arrays
             value = value.view("int64")
-        value = bytearray(memoryview(value))
+        value = bytes(memoryview(value))
     return value
 
 
 def get_mapper(
-    url,
+    url="",
     check=False,
     create=False,
     missing_exceptions=None,
